#!/usr/bin/perl
use strict;
use warnings;
use Cwd;

my $numPerms = 100;
my $topk = 2;
my $graphSize = 2;
my $firstSamSize = 5;
my $secondSamSize = 5;
my %reactZvalue = ();
my %reactZvalue3 = ();
my $depTag = 0;
my %rrDir = ();          #rrDir{r2}{r1}=left,right,all; r2 is at r1's which side
my %rneis = ();          #rneis{r1}{left, right}{r2}; r1 can go to r2 in an oriented manner
my %ubi = ubiCpd();
my %nullDistr = ();      #$nullDistr{m} = [array of values]; p2 null scores for size m subgraph
my %sigPathScore = ();         #$path2p{enr, dep}{path} = score
my %sigPathP = ();         #$path2p{enr, dep}{path} = [p1, p2]

sub setGlobeVars {
    ($firstSamSize, $numPerms) = @_;
}

#convert KO matrix to reaction matrix
sub koReactMatrix {
    my ($file) = @_;
    my %koReact = koReact();
    my $info = ""; #first line sample labels
    my %samReact = (); #samReact{react}{sam} = abundance
    open(FH, $file) or die("Could not open file: $file\n");
    foreach my $line (<FH>)
    {
	chomp $line;
	if ($line =~ /^\t/) #first line
	{
	    $info = $line;
	    next;
	}
	my ($kid, @samples) = split("\t", $line);
	for(my $i = 0; $i < scalar @samples; $i++)
	{
	    foreach my $react (keys %{$koReact{$kid}})
	    {
		$samReact{$react}{$i} += $samples[$i];
	    }
	}
    }
    close FH;
    open(OUTPUT, ">./tmpOutput/react.matrix");
    select OUTPUT;
    print $info, "\n";
    foreach my $react (sort keys %samReact)
    {
	print $react, "\t";
	my @nums = ();
	foreach my $sam (sort {$a <=> $b} keys %{$samReact{$react}})
	{
	    push(@nums, $samReact{$react}{$sam});
	}
	print join("\t", @nums);
	print "\n";
    }
    select STDOUT;
    close OUTPUT;
}

sub koReact {
    my $procdatadir = procDataDir();
    my %sameReacts = (); #store redundant reactions
    open(FH, "$procdatadir/sameReacts.tab");
    foreach my $line (<FH>) {
	chomp $line;
	if ($line =~ /^R/) {
	    my (@reacts) = split("\t", $line);
	    for (my $i = 1; $i < scalar @reacts; $i++) {
		$sameReacts{$reacts[$i]} = $reacts[0];
	    }
	}
    }
    my $koreactfile = "$procdatadir/koReact.tab";
    open(FH, $koreactfile);
    my %koreact = ();
    foreach my $line (<FH>) {
	chomp($line);
	if ($line =~ /^\s/) {next;} #empty lines
	my ($ko, $react) = split("\t", $line);
	if (exists $sameReacts{$react})	{ next;} #redundant reactions
	$koreact{$ko}{$react} = 1;
    }
    close FH;
    return %koreact;
}

sub computeZ
{
#    $topk = int (100 / $num);
#    print $topk;
    chdir("./src"); 
    print "Running Metastats on reaction matrix ... ";
    runMetastat();
    print "done!\n";
    print "Converting P values to Z values ... ";
    p2z();
    print "done!\n";
    chdir("../");
}

sub runMetastat
{
    my $num = $firstSamSize + 1;
    open(OUTPUT, ">metastat.r");
    select OUTPUT;
    print 'source("detect_DA_features.r")', "\n";
    print 'jobj <- load_frequency_matrix("../tmpOutput/react.matrix")', "\n";
    print 'detect_differentially_abundant_features(jobj, ', $num, ', "../tmpOutput/react.metastat")', "\n";
    select STDOUT;
    close OUTPUT;
    system(" R --slave --vanilla < metastat.r > ../tmpOutput/tem");
}

sub p2z
{
    my $samNum = $firstSamSize;
    ########################################
    #extract p values from raw metastat output
    open (OUTPUT, ">../tmpOutput/react2p.tab");
    open(FH, "../tmpOutput/react.metastat");
    foreach my $line (<FH>)
    {
	chomp $line;
	my ($react, $b, $c, $d, $e, $f, $g, $pvalue, $q) = split("\t", $line);
	$pvalue = $pvalue / 2;
	if ($pvalue == 1)
	{
	    print OUTPUT "$react\t0.999\t";
	}
	elsif ($pvalue < 0.001)
	{
	    print OUTPUT "$react\t0.001\t";
	}
	else
	{
	    print OUTPUT "$react\t$pvalue\t";
	}

	if ($q < 0.001)
	{
	    print OUTPUT "0.001\n";
	}
	else
	{
	    print OUTPUT "$q\n";
	}
    }
    close FH;
    close OUTPUT;
    ########################################

    ########################################
    #convert p to z using p2z.r
    system("R --slave --vanilla < p2z.r > ../tmpOutput/tem"); #output react2z.raw
    ########################################

    ########################################
    #compute the sign of z values, because metastats computes two tailed tests
    open(FH, "../tmpOutput/react.matrix");
    my @reactmatrix = <FH>;
    close FH;
    my @samSize = (); #sample size
    foreach my $line (@reactmatrix)
    {
	if ($line =~ /^\s/) { next;}
	my ($react, @values) = split("\t", $line);
	$secondSamSize = scalar @values - $firstSamSize;
	for (my $i = 0; $i < scalar @values; $i++)
	{
	    $samSize[$i] += $values[$i]; #compute sample size
	}
    }
    my %react2level  = (); #which population is more abundant for each reaction
    foreach my $line (@reactmatrix)
    {
	if ($line =~ /^\s/) { next;}
	my ($react, @values) = split("\t", $line);
	my $one = 0; #group 1
	for (my $i = 0; $i < $samNum; $i++) 
	{
	    $one += $values[$i] / $samSize[$i];
	}
	my $two = 0; #group 2
	for (my $i = $samNum; $i < scalar @values; $i++)
	{
	    $two += $values[$i] / $samSize[$i];
	}
	$react2level{$react} = $one/$firstSamSize - $two/$secondSamSize; #diff of average
    }
    ########################################

    ########################################
    #add sign to z values
    open(FH, "../tmpOutput/react2p.tab"); #get the ordered reacts
    my @reacts = ();
    foreach my $line (<FH>)
    {
	my ($react, $pvalue) = split("\t", $line);
	push(@reacts, $react); #all reacts in order
    }
    close FH;
    open(FH, "../tmpOutput/react2z.raw"); #read raw z values
    my @p2zraw = <FH>;
    close FH;
    my $i = 0;
    open(OUTPUT, ">../tmpOutput/react2zEnriched.tab");
    select OUTPUT;
    open(OUTPUT2, ">../tmpOutput/react2zDepleted.tab");
    foreach my $line (@p2zraw)
    {
	my (@values) = split(/\s+/, $line);
	foreach my $ele (@values)
	{
	    if ($react2level{$reacts[$i]} > 0) #group 1 more abundant
	    {
		print $reacts[$i], "\t$ele\n";
		print OUTPUT2 $reacts[$i], "\t", $ele * -1, "\n";
	    }
	    else #group 2 more abundant
	    {
		print $reacts[$i], "\t", $ele * -1, "\n";
		print OUTPUT2 $reacts[$i], "\t$ele\n";
	    }
	    $i++;
	}
    }
    select STDOUT;
    close OUTPUT;
    system("cp ../tmpOutput/react2zEnriched.tab ../tmpOutput/react2z.tab");
}

sub setDepleted
{
    $depTag = 1;
    system("cp ./tmpOutput/react2zDepleted.tab ./tmpOutput/react2z.tab");
}

sub react2zvalues {
    my $react2ztab = "./tmpOutput/react2z.tab";
    open(FH, $react2ztab);
    my @react2ztab = <FH>;
    close FH;
    foreach my $line (@react2ztab) {
	chomp $line;
	if ($line =~ /^\t/) { next;}           #blank line
	my ($react, $z) = split("\t", $line);
	$reactZvalue{$react} = $z;             #store in global variable
    }
}

sub searchMWS {
    react2zvalues();                            #parse react2z.tab, create reactZvalue
    rNeis();                                    #parse reaction_mapformula.lst, create rrDir and rneis
    my %reactZvalue2 = %reactZvalue;            #%reactZvalue will be changed every time
    %reactZvalue3 = %reactZvalue;               #%reactZvalue will be changed every time
    my %preReacts = ();                         #previous reacts during a single search
    my $j = 1;
    while (1) {
	print "Searching number $j maximum weighted subgraph ... ";
	my %graph2z = greedySearch();           #a set of MWSs starting from each node
	print "done!\n";
	my $reactStr = "";
	foreach my $str (sort {$graph2z{$b} <=> $graph2z{$a}} keys %graph2z) {
	    $reactStr = $str;
	    last;                               #only use the best max weight graph
	}
	my @reacts = split(/\$/, $reactStr);
	my $netsize = scalar @reacts;
	my $sumz = $graph2z{$reactStr};
	print join("\t", @reacts), "\tScore=$sumz\n";
	print "Computing p values for this subpathway ... ";
	my $p2 = computeP2($netsize, $sumz);    #compute p2 value
	my $p1 = computeP1($reactStr, $sumz);
	print "done!\n";
	printf("p1=%.3f\tp2=%.3f\n\n", $p1, $p2);
	my $numReacts = 0;
	foreach my $reactId (@reacts) {
	    $preReacts{$reactId} = 1;
	    delete $reactZvalue{$reactId};
	}
	$j++;
	if ($p2 > 0.05) { last;}                #nonsignificant
	if ($depTag) {
	    $sigPathScore{'Dep'}{$reactStr} = $sumz;
	    $sigPathP{'Dep'}{$reactStr} = [$p1, $p2];
	} else {
	    $sigPathScore{'Enr'}{$reactStr} = $sumz;
	    $sigPathP{'Enr'}{$reactStr} = [$p1, $p2];
	}

    }
    close FH;
    %reactZvalue = %reactZvalue2;
}

sub computeP2 {
    my ($netsize, $netz) = @_;
    if (exists $nullDistr{$netsize}) {}
    else {
	nullNetScore($netsize);
    }
    my $i = 0;
    #print "size$netsize\n";
    foreach my $value (@{$nullDistr{$netsize}}) {
	#print "$value\t";
	if ($netz <= $value) { $i++;}
    }
#    print "\n";
    return $i / $numPerms;
}

sub nullNetScore {
    my ($size) = @_;
    my %reactZvalue4 = %reactZvalue3;                         #%reactZvalue3 will be modified during search
    for (my $j = 0; $j < $numPerms; $j++) {
	my @zvalues = values %reactZvalue3;
	my @reacts = keys %reactZvalue3;
	randArray(\@zvalues);                                 #permute z values for each reaction
	for (my $i = 0; $i < scalar @reacts; $i++) {
	    $reactZvalue3{$reacts[$i]} = $zvalues[$i];
	}
	my %graph2z = greedySearchSize($size);                #find MWS with fixed size
	foreach my $reactStr (sort {$graph2z{$b} <=> $graph2z{$a}} keys %graph2z) {
	    my @reacts = split(/\$/, $reactStr);
	    my $z = $graph2z{$reactStr};
	    push(@{$nullDistr{"$size"}}, $z);
	    last;                                             #only use the top 1
	}
    }
    %reactZvalue3 = %reactZvalue4;                            #reinitialize %reactZvalue3
}

sub greedySearchSize {
    my ($size) = @_;
    my %graph2z = ();
    foreach my $react (sort keys %reactZvalue3) {                       #search from each react
	foreach my $dir (("right", "left")) {                           #for each direaction for each react
	    my %maxGraph = findNodeMaxGraphSize($react, $dir, $size);   #find MWS for each react
	    if (scalar keys %maxGraph != $size) { next;}                #size is not satisfied
	    my $graphZ = 0;
	    my $graphID = join('$', sort keys %maxGraph);
	    foreach my $react2 (keys %maxGraph) {
		$graphZ += $reactZvalue3{$react2};
	    }
	    $graphZ = $graphZ / sqrt(scalar keys %maxGraph);
	    $graph2z{$graphID} = $graphZ;
	}
    }
    return %graph2z;
}

sub findNodeMaxGraphSize {
    my ($react, $dir, $size) = @_;
    my %maxGraph = ($react => 1);
    for (my $i = 0; $i < $size - 1; $i++) {        #$size-1: first reaction already included
	my ($bestNei, $newdir) = findBestNeighbor($react, $dir, %maxGraph);
	$react = $bestNei;
	$dir = $newdir;
	if ($bestNei eq "") {                      #subgraph cannot expand anymore
	    return %maxGraph;
	}
	$maxGraph{$bestNei} = 1;
    }
    return %maxGraph;
}

#greedy search without size restriction
sub greedySearch {
    my %graph2z = ();
    foreach my $react (keys %reactZvalue) {                  #start from every react
	foreach my $dir (("right", "left")) {                #both left and right for each react
	    my %maxGraph = findNodeMaxGraph($react, $dir);   #start from $react with $dir direction, find MWS
	    if (scalar keys %maxGraph < 2) { next;}          #can not be extended
	    my $graphZ = 0;
	    my $graphID = join('$', sort keys %maxGraph);  
	    foreach my $react2 (keys %maxGraph) {
		$graphZ += $reactZvalue{$react2};
	    }
	    $graphZ = $graphZ / sqrt(scalar keys %maxGraph); #normalized Z value
	    $graph2z{$graphID} = $graphZ;
	}
    }
    return %graph2z;
}

sub findNodeMaxGraph {
    my ($react, $dir) = @_;
    my %maxGraph = ($react => 1);             #contains first reaction
    my $tag = 0;
    my $sumz = $reactZvalue{$react};          #z value for maxGraph
    my $pres = $reactZvalue{$react};          #z value for smaller maxGraph during search
    my $iter = 1;
    my %tem = ();
    while ($tag < 3) {                        #z score has been reduced less than 3 times
	$iter++;                              #graph size
	my ($bestNei, $newdir) = findBestNeighbor($react, $dir, %maxGraph);
	$dir = $newdir;
	$react = $bestNei;
	if ($bestNei eq "") { last;}	      #subgraph cannot expand anymore
	$sumz += $reactZvalue{$bestNei};
	my $score = $sumz / sqrt($iter);      #normalized graph score
	$maxGraph{$bestNei} = 1;
	if ($score > $pres) {                 #sumz is not reduced
	    %tem = ();                        #score-reducing reactions
	    $tag = 0;
	    $pres = $score;
	} else {
	    $tag++;                           #sumz is reduced
	    $tem{$bestNei} = 1;               #potentially needs to be removed
	    next;
	}
    }
    #more than two score-reducing reactions in a row
    foreach my $ele (keys %tem) {             #remove the last two score-reducing reactions
	delete $maxGraph{$ele};
    }
    return %maxGraph;
}

sub findBestNeighbor {
    my ($react, $dir, %reacts) = @_;
    my $bestreact = "";
    my $bestdir = "";
    my $maxz = -100;
    if (exists $rneis{$react}{$dir}) {                          #react can be extended in $dir direction
	foreach my $nei (keys %{$rneis{$react}{$dir}}) {
	    if (exists $reacts{$nei}) { next;}                  #already in current subpathway
	    unless (exists $reactZvalue{$nei}) { next;}
	    if ($rrDir{$react}{$nei} eq "left") {               #connects at $nei's left
		if ($reactZvalue{$nei} > $maxz &&     		#has highest z value in neighbors
		    exists $rneis{$nei}{"right"}) {		#and can be extended further
		    $maxz = $reactZvalue{$nei};
		    $bestreact = $nei;
		    $bestdir = "right";
		}
	    } else {                                            #connects at $nei's right
		if ($reactZvalue{$nei} > $maxz &&
		    exists $rneis{$nei}{"left"}) {
		    $maxz = $reactZvalue{$nei};
		    $bestreact = $nei;
		    $bestdir = "left";
		}
	    }
	}
    }
    return ($bestreact, $bestdir);
}

sub procDataDir
{
    return "./data/";
}

sub ubiCpd
{
    my $procdatadir = procDataDir();
    my $ubicpdfile = "$procdatadir/ubiCpd.tab";
    open(FH, $ubicpdfile);
    my @data = <FH>;
    close FH;
    my %ubiCpd = ();
    foreach my $line (@data)
    {
	chomp($line);
	if ($line =~ /^\s/) { next;}
	elsif ($line =~ /^cpdID/) { next;}
	else
	{
	    my ($cpd, $name) = split("\t", $line);
	    $ubiCpd{$cpd} = $name;
	}
    }
    return %ubiCpd;
}

sub randArray {
    my $array = shift;
    my $i;
    srand;
    for ($i = @$array; --$i; ) {
        my $j = int rand ($i+1);
        next if $i == $j;
        @$array[$i,$j] = @$array[$j,$i];
    }
}

sub computeP1 {
    my ($reactStr, $zscore) = @_;
    open(FH, "./tmpOutput/react.matrix");
    my @reactmatrix = <FH>;
    close FH;
    my @samSize = ();
    my %react2nums = ();
    my $label = "";
    foreach my $line (@reactmatrix) {
	chomp $line;
	if ($line =~ /^\s/) {
	    $label = $line;
	    next;
	}
	my ($react, @values) = split("\t", $line);
	for (my $i = 0; $i < scalar @values; $i++) { #compute sample size
	    $samSize[$i] += $values[$i]; 
	}
	$react2nums{$react} = [@values];
    }

    my @samSizeTmp = @samSize;
    my @reacts = split(/\$/, $reactStr);
    foreach my $react (@reacts) {
	for (my $i = 0; $i < scalar @{$react2nums{$react}}; $i++) {
	    $samSizeTmp[$i] -= $react2nums{$react} -> [$i];
	}
    }
    my $samnum = $firstSamSize + $secondSamSize;
    my @dumary = ();
    for (my $i = 0; $i < $samnum; $i++) {
	$dumary[$i] = $i;
    }

    for (my $i = 0; $i < $numPerms; $i++) {
	open(OUTPUT, ">./tmpOutput/perm/perm$i.matrix");
	select OUTPUT;
	print $label, "\n";
	randArray(\@dumary);
	foreach my $react (@reacts) {
	    my @nums = @{$react2nums{$react}};
	    my @numsPerm;
	    for (my $j = 0; $j < $samnum; $j++) {
		my $index = $dumary[$j];
		push(@numsPerm, $nums[$index]);
	    }
	    print $react, "\t", join("\t", @numsPerm), "\n";
	}
	my @samSizeTmpPerm;
	for (my $j = 0; $j < $samnum; $j++) {
	    my $index = $dumary[$j];
	    push(@samSizeTmpPerm, $samSizeTmp[$index]);
	}
	print "Other\t", join("\t", @samSizeTmpPerm), "\n";
	select STDOUT;
	close OUTPUT;
    }
    metastat1000();
    chdir("./src");
    system(" R --slave --vanilla < metastat1000.r > ../tmpOutput/tem");
    my @values = p2z1000();
    chdir("../");
    my $k = 0;
    foreach my $value (@values) {
	if ($value >= $zscore) {
	    $k++;
	}
    }
    return $k / $numPerms;
}

sub p2z1000 {
    my $samNum = $firstSamSize;
    open (OUTPUT, ">../tmpOutput/permMetastat1000.tab");
    for (my $i = 0; $i < $numPerms; $i++) {
	open(FH, "../tmpOutput/permMetastat/perm$i.metastat");
	my @metastatRaw = <FH>;
	close FH;
	foreach my $line (@metastatRaw)	{
	    chomp $line;
	    my ($react, $b, $c, $d, $e, $f, $g, $pvalue, $i) = split("\t", $line);
	    $pvalue = $pvalue / 2;
	    if ($pvalue == 1) {
		print OUTPUT "$react\t0.999\n";
	    } elsif ($pvalue < 0.001) {
		print OUTPUT "$react\t0.001\n";
	    } else {
		print OUTPUT "$react\t$pvalue\n";
	    }
	}
    }
    close OUTPUT;
    system("R --slave --vanilla < p2zPerm.r > ../tmpOutput/tem"); #output react2z.raw
    open(FH, "../tmpOutput/permMetastat1000z.raw"); #read raw z values
    my @p2zraw = <FH>;
    close FH;
    my @allz = ();
    foreach my $line (@p2zraw) {
	my (@values) = split(/\s+/, $line);
	foreach my $ele (@values) {
	    push(@allz, $ele);
	}
    }
    my $tag = 0;
    my @zvalues = ();
    for (my $i = 0; $i < $numPerms; $i++) {
	open(FH, "../tmpOutput/perm/perm$i.matrix");
	my @reactmatrix = <FH>;
	close FH;
	my @samSize = ();
	foreach my $line (@reactmatrix) {
	    if ($line =~ /^\s/) { next;}
	    my ($react, @values) = split("\t", $line);
	    for (my $i = 0; $i < scalar @values; $i++) {
		$samSize[$i] += $values[$i]; #compute sample size
	    }
	}
	my %react2level  = ();
	my @levels = ();
	foreach my $line (@reactmatrix)	{
	    if ($line =~ /^\s/) { next;}
	    my ($react, @values) = split("\t", $line);
	    my $one = 0; #group 1
	    for (my $i = 0; $i < $samNum; $i++) {
		$one += $values[$i] / $samSize[$i];
	    }
	    my $two = 0; #group 2
	    for (my $i = $samNum; $i < scalar @values; $i++) {
		$two += $values[$i] / $samSize[$i];
	    }
	    my $level = $one/$firstSamSize - $two/$secondSamSize;
	    $react2level{$react} = $level;
	    push(@levels, $level);
	}
	foreach my $level (@levels) {
	    if ($level > 0) {
		if ($depTag) {
		    push(@zvalues, $allz[$tag] * -1);
		} else {
		    push(@zvalues, $allz[$tag] * 1);
		}
	    } else {
		if ($depTag) {
		    push(@zvalues, $allz[$tag] * 1);
		} else {
		    push(@zvalues, $allz[$tag] * -1);		    
		}
	    }
	    $tag++;
	}
    }
    my @permValues = ();
    for (my $i = 0; $i < $numPerms; $i++) {
	my $sum = 0;
	for (my $j = 0; $j < $graphSize; $j++) {
	    $sum += $zvalues[$i * $graphSize + $j];
	}
	push(@permValues, $sum);
    }
    return @permValues;
}


sub metastat1000
{
    open(OUTPUT, ">./src/metastat1000.r");
    select OUTPUT;
    print 'source("detect_DA_features.r")', "\n";
    for (my $i = 0; $i < $numPerms; $i++) {
	print 'jobj <- load_frequency_matrix("../tmpOutput/perm/perm', $i, '.matrix")', "\n";
	print 'detect_differentially_abundant_features(jobj, ', $firstSamSize + 1, ', "../tmpOutput/permMetastat/perm', $i, '.metastat", B=100)', "\n\n";
    }
    select STDOUT;
    close OUTPUT;
}

sub permP
{
    my ($value, @values) = @_;
    my $num = 0;
    foreach my $ele (@values)
    {
	if ($ele > $value)
	{
	    $num++;
	}
    }
    if ($num == 0)
    {
	return 1 / scalar @values;
    }
    else
    {
	return $num / (scalar @values);
    }
}

sub outputResults {
    my %enrreacts = (); 
    open(FH, "./tmpOutput/react2zEnriched.tab");
    foreach my $line (<FH>) {
	chomp $line;
	my ($react, $z) = split("\t", $line);
	if ($z > 0) {
	    $enrreacts{$react} = 1;
	}
    }

    my %react2p = ();
    my %react2q = ();
    open(FH, "./tmpOutput/react2p.tab");
    foreach my $line (<FH>) {
	chomp $line;
	my ($react, $p, $q) = split("\t", $line);
	if ($p <= 0.05)	{
	    if (exists $enrreacts{$react}) {
		$react2p{"Enr"}{$react} = $p;
		$react2q{"Enr"}{$react} = $q;
	    } else {
		$react2p{"Dep"}{$react} = $p;
		$react2q{"Dep"}{$react} = $q;
	    }
	}
    }
    
    my %printedReacts = ();
    my @file = ("Enr", "Dep");
    foreach my $file (@file) {
	if ($file eq "Enr") {
	    print "Significant subpathwyas enriched in the first population\n";
	} else {
	    print "\nSignificant subpathwyas enriched in the second population\n";
	}
	print "p1\tp2\tSignificant Subpathway\n";
	foreach my $path (sort {$sigPathScore{$file}{$b} <=> $sigPathScore{$file}{$a}} keys %{$sigPathScore{$file}}) {
	    printf("%.3f\t%.3f\t", $sigPathP{$file}{$path} -> [0], $sigPathP{$file}{$path} -> [1]);
	    my @reacts = split(/\$/, $path);
	    print join("\t", @reacts), "\n";
	    foreach my $react (@reacts) {
		$printedReacts{$react} = 1;
	    }
	}

    }
    
    foreach my $file (@file) {
	if ($file eq "Enr") {
	    print "\nAdditional Significant Reactions Enriched in First Population\n";
	} else {
	    print "\nAdditional Significant Reactions Enriched in Second Population\n";
	}
	print "p1\tp2\treaction\n";
	foreach my $react (sort {$react2p{$file}{$a} <=> $react2p{$file}{$b}}keys %{$react2p{$file}}) {
	    if (exists $printedReacts{$react}) { next;}
	    printf("%0.3f\t", $react2p{$file}{$react});
	    printf("%0.3f\t", $react2q{$file}{$react});
	    print "$react\n";
	}
    }
}

sub rNeis {
    my %reactDir = ();     #direction (left, right or all) of each reaction
    my %cpdReactOri = ();  #side (left or right) of each molecule in each direction
    open(FH, "./data/reaction_mapformula.lst");
    my @reactdata = <FH>;
    close FH;
    foreach my $line (@reactdata) {
	chomp $line;
	$line =~ /^(R\d+)\:\s(\d+)\:\s(.*)$/;
	my ($rid, $pid, $react) = ($1, $2, $3);
	unless (exists $reactZvalue{$rid}) { next;}     #ignore if not present in the reaction abundance matrix
	if ($react =~ /<=>/) {                          #reaction can go both directions
	    my ($left, $right) = split("<=>", $react);
	    my @lefts = split(/\+/, $left);             #left-side molecules
	    my @rights = split(/\+/, $right);           #right-side molecules
	    foreach my $l (@lefts) {                    
		$l =~ /(\S+)/;
		$l = $1;
		$cpdReactOri{$l}{$rid} = "left";        #molecule is at reaction's left side
	    }
	    foreach my $r (@rights) {
		$r =~ /(\S+)/;
		$r = $1;
		$cpdReactOri{$r}{$rid} = "right";
	    }
	    $reactDir{$rid} = "all";
	} elsif ($react =~ /=>/) {                      #reaction can go right
	    my ($left, $right) = split("=>", $react);
	    my @lefts = split(/\+/, $left);
	    my @rights = split(/\+/, $right);
	    foreach my $l (@lefts) {
		$l =~ /(\S+)/;
		$l = $1;
		$cpdReactOri{$l}{$rid} = "left";
	    }
	    foreach my $r (@rights) {
		$r =~ /(\S+)/;
		$r = $1;
		$cpdReactOri{$r}{$rid} = "right";
	    }
	    if (exists $reactDir{$rid}) {               #this reaction already has direction
		if ($reactDir{$rid} eq "right")	{}      #same direction as this reaction(right)
		else {                                  #left or all
		    $reactDir{$rid} = "all";            #change direction to all
		}
	    } else {
		$reactDir{$rid} = "right";              #this reaction has been seen before
	    }
	} elsif ($react =~ /<=/) {                      #reaction can go left
	    my ($left, $right) = split("<=", $react);
	    my @lefts = split(/\+/, $left);
	    my @rights = split(/\+/, $right);
	    foreach my $l (@lefts) {
		$l =~ /(\S+)/;
		$l = $1;
		$cpdReactOri{$l}{$rid} = "left";
	    }
	    foreach my $r (@rights) {
		$r =~ /(\S+)/;
		$r = $1;
		$cpdReactOri{$r}{$rid} = "right";
	    }
	    if (exists $reactDir{$rid}) {
		if ($reactDir{$rid} eq "left") {}
		else {
		    $reactDir{$rid} = "all";
		}
	    } else {
		$reactDir{$rid} = "left";
	    }
	}#end of parsing each reaction
    }#end of parsing the whole file

    #allow two reactions to be neighbors if sharing molecules
    foreach my $cpd (sort keys %cpdReactOri) {                 #for each molecule
	if (exists $ubi{$cpd}) { next;}                        #ubiquitous molecule
	foreach my $react1 (keys %{$cpdReactOri{$cpd}})	{      #for each associated reaction
	    my $dir = $cpdReactOri{$cpd}{$react1};             #direction of this molecule in reaction
	    foreach my $react2 (keys %{$cpdReactOri{$cpd}}) {  #for each associated reaction
		if ($react1 eq $react2) { next;}                     #same reaction
		$rrDir{$react2}{$react1} = $dir;               #react2 is at react1's $dir side
	    }
	}
    }

    #compute oriented neighbors
    foreach my $cpd (sort keys %cpdReactOri) {
	if (exists $ubi{$cpd}) { next;}
	foreach my $react1 (keys %{$cpdReactOri{$cpd}})	{
	    my $dir1 = $cpdReactOri{$cpd}{$react1};
	    if ($dir1 eq "right") {                                       # = cpd
		if ($reactDir{$react1} eq "left") { next;}                # <= cpd, react can NOT be extended through this molecule
		else {                                                    # => cpd or <=> cpd, can be extended
		    foreach my $react2 (keys %{$cpdReactOri{$cpd}}) {
			if ($react1 eq $react2) { next;}                  #same reaction
			my $dir2 = $cpdReactOri{$cpd}{$react2};
			if ($dir2 eq "left") {                            # =>r1 cpd =r2
			    if ($reactDir{$react2} eq "left") { next;}    # =>r1 cpd <=r2
			    else {                                        # =>r1 cpd =>r2
				$rneis{$react1}{"right"}{$react2} = 1;    #react1 goes to react2 through right molecule
			    }
			} else {                                          # =>r1 cpd =r2 cpd
			    if ($reactDir{$react2} eq "right") { next;}   # =>r1 cpd =>r2 cpd
			    else {                                        # =>r1 cpd <=r2 cpd OR =>r1 cpd <=>r2 cpd
				$rneis{$react1}{"right"}{$react2} = 1;
			    }
			}
		    }
		}
	    } else {                                                      # cpd =
		if ($reactDir{$react1} eq "right") { next;}               # cpd =>
		else {                                                    # cpd <= or cpd <=>
		    foreach my $react2 (keys %{$cpdReactOri{$cpd}}) {
			if ($react1 eq $react2) { next;}                  #same reaction
			my $dir2 = $cpdReactOri{$cpd}{$react2};
			if ($dir2 eq "right") {                           # =r2 cpd <=r1
			    if ($reactDir{$react2} eq "right") { next;}   # =>r2 cpd <=r1
			    else {                                        # <=r2 cpd <=r1
				$rneis{$react1}{"left"}{$react2} = 1;
			    }
			} else {                                          # cpd =r2 cpd <= r1
			    if ($reactDir{$react2} eq "left") { next;}    # cpd <=r2 cpd <= r1
			    else {                                        # cpd =>r2 cpd <= r1
				$rneis{$react1}{"left"}{$react2} = 1;
			    }
			}
		    }
		}
	    }
	}
    }#end of computing oriented neighbors
}

sub cleanTmpFiles
{
    system("rm ./tmpOutput/path*");
    system("rm ./tmpOutput/perm/*");
    system("rm ./tmpOutput/permMetastat/*");
}

1;
