#!/usr/bin/perl

#######################################
#
# Program: MetaPath uses high-throughput metagenomic sequences
#          and prior metabolic pathways knowledge to identify
#          differentially abundant subpathways.
#
# Author:  Bo Liu, Graduate Student
#          Department of Computer Science,
#          Center for Bioinformatics and Computational Biology,
#          University of Maryland, College Park
#
# Email:   boliu@umiacs.umd.edu
#
# Last Modified: Feb 16, 2010
#
#######################################

use strict;
use warnings;
use libMetaPath;
use Getopt::Std;

my $kofile = "";       #input ko matrix file
my $samNum = 0;        #number of samples in first population
my $numPerms = 100;    #number of permutations
my $outputfile = "";   #output file 
my %opts = ();
getopts('k:n:p:o:', \%opts);
#mandatory option that must be specified
if (exists $opts{"k"} && exists $opts{"n"}) {
    $kofile = $opts{"k"}; 
    $samNum = $opts{"n"}; 
} else {
    Usage();
    exit;
}
if (exists $opts{"p"}) {
    $numPerms = $opts{"p"};
}
if (exists $opts{"o"}) {
    $outputfile = $opts{"o"};
}

print "\nConverting input KO matrix to reaction matrix ... ";
koReactMatrix($kofile);
print "done!\n";

setGlobeVars($samNum, $numPerms);

#compute z values for each reaction
computeZ(); 

print "\n";
print "######################################\n";
print "Finding Significantly Enriched Subpathways in the First Population\n";
print "######################################\n";
searchMWS();

print "\n";
print "######################################\n";
print "Finding Significantly Enriched Subpathways in the Second Population\n";
print "######################################\n";
setDepleted();
searchMWS();
if ($outputfile)
{
    open(OUTPUT, ">$outputfile") or die("Can not open $outputfile\n");
    select OUTPUT;
}
print "\n";
print "######################################\n";
print "Results\n";
print "######################################\n";
outputResults();
exit;
if ($outputfile)
{
    close OUTPUT;
    select STDOUT;
}
print "\n";
print "Deleting temporary files ... ";
#cleanTmpFiles();
print "done!\n";
print "\n";
exit;

sub Usage {
    print "\nUsage:\n";
    print "\tperl metaPath.pl -k [KO file] -n [sample size of first group]\n";
    print "\nOption:\n";
    
    print "\t-p\tnumber of permutations\n";
    print "\t\tdefault=100\n\n";

    print "\t-o\toutput file\n";
    print "\t\tdefault: print on the screen\n\n";

    print "\nContact:\n";
    print "\t", 'Still have problems? Contact boliu@umiacs.umd.edu', "\n";
    print "\n";
}
